/*
  tsv2ics.cc

  Copyright (C) 2011 Taiji Yamada <taiji@aihara.co.jp>
 */
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <boost/optional.hpp>
#include "date_time_numeric.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#if defined(USE_BOOST_XPRESSIVE)
#include "boost_xpressive_utility.hpp"
using namespace boost::xpressive;
typedef string_sregex_map string_regex_map;
#else
#include "boost_regex_utility.hpp"
#endif
#include "repeater.hpp"
#include "date_time_repeater.hpp"
using namespace std;
using namespace boost;

int usage(ostream &os, int argc, char *argv[])
{
  (void)argc;
  os << "usage:\t" << argv[0] << " [--ol2k|ical] [--] -|filename.tsv [...]" << endl;
  return os == cerr ? 1 : 0;
}

int main(int argc, char *argv[])
{
  const char *vcalendar_fmt_ol2k[] = {
    "BEGIN:VCALENDAR\r\n\
PRODID:-//Taiji Yamada <taiji@aihara.co.jp>//tsv2ics//EN\r\n\
METHOD:PUBLISH\r\n\
VERSION:2.0\r\n\
BEGIN:VTIMEZONE\r\n\
TZID:Japan\r\n\
BEGIN:STANDARD\r\n\
DTSTART:19390101T000000\r\n\
TZOFFSETFROM:+0900\r\n\
TZOFFSETTO:+0900\r\n\
TZNAME:JST\r\n\
END:STANDARD\r\n\
END:VTIMEZONE\r\n\
",
    "END:VCALENDAR\r\n"
  }, *vevent_fmt_ol2k[] = {
    "BEGIN:VEVENT\r\n\
DTSTAMP:%1%\r\n\
SUMMARY:%2%\r\n\
LOCATION:%3%\r\n\
DESCRIPTION:%4%\r\n\
DTSTART%5%\r\n\
DTEND%6%\r\n\
CATEGORIES:%7%\r\n\
TRANSP:%8%\r\n\
PRIORITY:0\r\n\
CLASS:PUBLIC\r\n",
    "END:VEVENT\r\n",
    ";TZID=Japan:%1%T000000",
    ";TZID=Japan:%1%"
  }, *vcalendar_fmt_ical[] = {
    "BEGIN:VCALENDAR\r\n\
PRODID:-//Taiji Yamada <taiji@aihara.co.jp>//tsv2ics//EN\r\n\
METHOD:PUBLISH\r\n\
VERSION:2.0\r\n\
X-WR-TIMEZONE:Asia/Tokyo\r\n\
CALSCALE:GREGORIAN\r\n\
",
    "END:VCALENDAR\r\n"
  }, *vevent_fmt_ical[] = {
    "BEGIN:VEVENT\r\n\
DTSTAMP:%1%\r\n\
SUMMARY:%2%\r\n\
LOCATION:%3%\r\n\
DESCRIPTION:%4%\r\n\
DTSTART%5%\r\n\
DTEND%6%\r\n\
CATEGORIES:%7%\r\n\
TRANSP:%8%\r\n\
PRIORITY:0\r\n\
CLASS:PUBLIC\r\n",
    "END:VEVENT\r\n",
    ";VALUE=DATE:%1%",
    ";TZID=Asia/Tokyo:%1%"
  };
  const char **vcalendar_fmt = vcalendar_fmt_ol2k, **vevent_fmt = vevent_fmt_ol2k;
  string default_summary[2], default_description[2], default_location[2], default_categories, default_transp("TRANSPARENT");
  boost::optional<gregorian::date> force_limited_since, force_limited_until;
  bool normalize_until = false, convert_until_count = false, force_unlimited_repeat = false;

  enum tsv_type {
    tsv_date,
    tsv_year_month_days,
    tsv_month_day_years,
  } tsv_kind = tsv_month_day_years;
  streambuf *cin_rdbuf = cin.rdbuf();
  vector<string> filenames;
  int a;
  for (a=1; a<argc; ++a)
    if (string(argv[a]) == "--" && ++a)
      break;
    else if (string(argv[a]) == "--ol2k")
      vcalendar_fmt = vcalendar_fmt_ol2k, vevent_fmt = vevent_fmt_ol2k;
    else if (string(argv[a]) == "--ical")
      vcalendar_fmt = vcalendar_fmt_ical, vevent_fmt = vevent_fmt_ical;
    else if (string(argv[a]) == "--force_limited_since" && a+1 < argc)
      force_limited_since = boost::optional<gregorian::date>(gregorian::from_string(argv[++a]));
    else if (string(argv[a]) == "--force_limited_until" && a+1 < argc)
      force_limited_until = boost::optional<gregorian::date>(gregorian::from_string(argv[++a]));
    else if (string(argv[a]) == "--normalize_until")
      normalize_until = true;
    else if (string(argv[a]) == "--convert_until_count")
      convert_until_count = true;
    else if (string(argv[a]) == "--force_unlimited_repeat")
      force_unlimited_repeat = true;
    else
      filenames.push_back(argv[a]);

  if (!filenames.size())
    return usage(cerr, argc, argv);

  posix_time::ptime dt(posix_time::second_clock::local_time());

  cout << format(vcalendar_fmt[0]);
  string_regex_map regex_map;
  for (size_t i=0; i<filenames.size(); ++i) {
    string sp;
    ifstream ifs(filenames[i].c_str());
    cin.rdbuf((filenames[i] != "-") ? ifs.rdbuf() : cin_rdbuf);
    string sb;
    while (getline(cin, sb)) {
      smatch m;
      if (regex_match(sb, m, regex_map.at("^#DATE", regex_constants::icase))) {
	tsv_kind = tsv_date;
	continue;
      }
      else if (regex_match(sb, m, regex_map.at("^#YEAR(?:/|,?\\t+)MONTH(?:/|,?\\t+)DAYS(?:=([^\\t,]+))?(?:,([^\\t,]+))?(?:,([^\\t,]+))?(?:,?\\t+DAYS(?:=([^\\t,]+))?(?:,([^\\t,]+))?(?:,([^\\t,]+))?)?(?:,?\\t+CATEGORIES=([^\\t]*))?"))) {
	tsv_kind = tsv_year_month_days;
	if (m[1].str() != "") default_summary[0]	= m[1].str();
	if (m[2].str() != "") default_description[0]	= m[2].str();
	if (m[3].str() != "") default_location[0]	= m[3].str();
	if (m[4].str() != "") default_summary[1]	= m[4].str();
	if (m[5].str() != "") default_description[1]	= m[5].str();
	if (m[6].str() != "") default_location[1]	= m[6].str();
	if (m[7].str() != "") default_categories	= m[7].str();
	continue;
      }
      else if (regex_match(sb, m, regex_map.at("^#MONTH(?:/|,?\\t+)DAY(?:/|,?\\t+)YEARS(?:,?\\t+SUMMARY=([^\\t]*)(?:,?\\t+DESCRIPTION=([^\\t]*)(?:,?\\t+LOCATION=([^\\t]*)(?:,?\\t+CATEGORIES=([^\\t]*))?)?)?)?"))) {
	tsv_kind = tsv_month_day_years;
	if (m[1].str() != "") default_summary[0]	= m[1].str();
	if (m[2].str() != "") default_description[0]	= m[2].str();
	if (m[3].str() != "") default_location[0]	= m[3].str();
	if (m[4].str() != "") default_categories	= m[4].str();
	continue;
      }
      else if (regex_match(sb, m, regex_map.at("^#"))) {
	continue;
      }
      sb = regex_replace(sb, regex_map.at("\\s*#.*$"), "");
      switch (tsv_kind) {
      case tsv_date:
	{
	  posix_time::ptime dt0, dt1;
	  string dtstart, dtend, rrule, summary, description, location, categories, transp;
	  if (!regex_match(sb, m, regex_map.at("^([0-9T]+)(?:-([0-9T]+))?(?:\\t+RRULE[:=]([^\\t]*))?(?:\\t+SUMMARY[:=]([^\\t]*))?(?:\\t+DESCRIPTION[:=]([^\\t]*))?(?:\\t+LOCATION[:=]([^\\t]*))?(?:\\t+CATEGORIES[:=]([^\\t]*))?(?:\\t+TRANSP[:=]([^\\t]*))?", regex_constants::icase))) continue;
	  if (m[1].str() != "") dtstart		= m[1].str();
	  if (m[2].str() != "") dtend		= m[2].str();
	  if (m[3].str() != "") rrule		= m[3].str();
	  if (m[4].str() != "") summary		= m[4].str();
	  if (m[5].str() != "") description	= m[5].str();
	  if (m[6].str() != "") location	= m[6].str();
	  if (m[7].str() != "") categories	= m[7].str();
	  if (m[8].str() != "") transp		= m[8].str();
	  if (regex_match(dtstart, m, regex_map.at("(\\d{8})T(\\d{6})"))) {
	    string date0, time0, date1, time1;
	    date0 = m[1].str();
	    time0 = m[2].str();
	    if (dtend == "") {
	      date1 = date0;
	      time1 = time0;
	    }
	    else if (regex_match(dtend, m, regex_map.at("(\\d{8})T(\\d{6})"))) {
	      date1 = m[1].str();
	      time1 = m[2].str();
	    }
	    else if (regex_match(dtend, m, regex_map.at("(\\d{6})"))) {
	      date1 = date0;
	      time1 = m[1].str();
	    }
	    dt0 = posix_time::from_iso_string(date0 + "T" + time0);
	    dt1 = posix_time::from_iso_string(date1 + "T" + time1);
	  }
	  else if (regex_match(dtstart, m, regex_map.at("(\\d{8})"))) {
	    string date0, date1;
	    date0 = m[1].str();
	    if (dtend == "") {
	      gregorian::date dt0 = gregorian::from_undelimited_string(date0);
	      date1 = gregorian::to_iso_string(dt0 + gregorian::date_duration(1));
	    }
	    else if (regex_match(dtend, m, regex_map.at("(\\d{8})"))) {
	      date1 = m[1].str();
	    }
	    dt0 = posix_time::ptime(gregorian::from_undelimited_string(date0));
	    dt1 = posix_time::ptime(gregorian::from_undelimited_string(date1));
	  }
	  ostringstream os;
	  os << format(vevent_fmt[0]) % to_iso_string(dt) %
	    summary %
	    location %
	    description %
	    ((dt0.time_of_day() == posix_time::time_duration()) ? format(vevent_fmt[2]) % gregorian::to_iso_string(dt0.date()) : format(vevent_fmt[3]) % posix_time::to_iso_string(dt0)) %
	    ((dt1.time_of_day() == posix_time::time_duration()) ? format(vevent_fmt[2]) % gregorian::to_iso_string(dt1.date()) : format(vevent_fmt[3]) % posix_time::to_iso_string(dt1)) %
	    categories %
	    transp;
	  if (rrule != "") {
	    os << format("RRULE:%1%") % rrule;
	    os << "\r\n";
	  }
	  string s = regex_replace(os.str(), regex_map["\\\\r\\\\n"], "\r\n");
	  cout << s << format(vevent_fmt[1]);
	}
	break;
      case tsv_year_month_days:
	{
	  string summary, description;
	  if (!regex_match(sb, m, regex_map.at("^(\\d+)(?:/|,?\\t+)(\\d+)(?:/|,?\\t+)([.~\\d,]+)(?:,?\\t+([.~\\d,]+))?"))) continue;
	  int year = lexical_cast<int>(m[1].str()), month = lexical_cast<int>(m[2].str());
	  for (int k=0; k<2; ++k) {
	    vector<string> days_list;
	    string days(m[3+k].str());
	    sregex_iterator it(days.begin(), days.end(), regex_map.at("([.~\\d]+),?")), end;
	    for (; it!=end; ++it)
	      days_list.push_back((*it)[1].str());
	    sort(days_list.begin(), days_list.end());
	    for (size_t i=0; i<days_list.size(); ++i) {
	      repeater<int, int> day = lexical_cast<repeater<int, int> >(days_list[i]);
	      repeater<gregorian::date, gregorian::days> ymd;
	      boost::optional<gregorian::date> after, before;
	      if (day.is_repeat()) {
		ymd = repeater<gregorian::date, gregorian::days>(gregorian::date(year, month, day.front),
								 gregorian::days(day.diff),
								 gregorian::date(year, month, day.back));
		if (force_limited_until)
		  if (!(before = ymd.before(*force_limited_until)))
		    continue;
		if (force_limited_since)
		  if (!(after = ymd.after(*force_limited_since)))
		    continue;
		  else {
		    ymd.front = *after;
		    day.front = ymd.front.day();
		  }
	      }
	      gregorian::date dt0(year, month, day.front), dt1 = dt0 + gregorian::date_duration(1);
	      ostringstream os;
	      os << format(vevent_fmt[0]) % to_iso_string(dt) %
		default_summary[k] %
		default_location[k] %
		default_description[k] %
		(format(vevent_fmt[2]) % to_iso_string(dt0)) % (format(vevent_fmt[2]) % to_iso_string(dt1)) %
		default_categories %
		default_transp;
	      if (day.is_repeat()) {
		if (force_limited_until)
		  if (before) {
		    ymd.back = *force_limited_until;
		    if (normalize_until)
		      ymd.back = *before;
		    day.back = ymd.back.day();
		  }
		os << format("RRULE:FREQ=DAYLY;INTERVAL=%1%") % day.diff;
		if (convert_until_count && !day.is_unlimited())
		  os << format(";COUNT=%1%") % day.count();
		else if (!force_unlimited_repeat && !day.is_unlimited())
		  os << format(";UNTIL=%1%T000000Z") %
		    (lexical_cast<string>(year) + lexical_cast<string>(month) +
		     lexical_cast<string>(normalize_until ? day.last() : day.back));
		os << "\r\n";
	      }
	      string s = regex_replace(os.str(), regex_map["\\\\r\\\\n"], "\r\n");
	      cout << s << format(vevent_fmt[1]);
	    }
	  }
	}
	break;
      case tsv_month_day_years:
      default:
	{
	  string summary, description, location;
	  if (!regex_match(sb, m, regex_map.at("^(\\d{1,2})(?:/|,?\\t+)(\\d{1,2})(?:/|,?\\t+)([.~\\d,]+)(?:,?\\t+([^\\t,]+)(?:,?\\t+([^\\t,]+)(?:,?\\t+([^\\t,]+))?)?)?"))) continue;
	  int month = lexical_cast<int>(m[1].str()), day = lexical_cast<int>(m[2].str());
	  if (m[4].str() != "") summary = m[4].str();
	  if (m[5].str() != "") description = m[5].str();
	  if (m[6].str() != "") location = m[6].str();
	  int k=0;
	  {
	    vector<string> years_list;
	    string years(m[3+k].str());
	    sregex_iterator it(years.begin(), years.end(), regex_map.at("([.~\\d]+),?")), end;
	    for (; it!=end; ++it)
	      years_list.push_back((*it)[1].str());
	    sort(years_list.begin(), years_list.end());
	    for (size_t i=0; i<years_list.size(); ++i) {
	      repeater<int, int> year = lexical_cast<repeater<int, int> >(years_list[i]);
	      repeater<gregorian::date, gregorian::years> ymd;
	      boost::optional<gregorian::date> after, before;
	      if (year.is_repeat()) {
		ymd = repeater<gregorian::date, gregorian::years>(gregorian::date(year.front, month, day),
								  gregorian::years(year.diff),
								  gregorian::date(year.back, month, day));
		if (force_limited_until)
		  if (!(before = ymd.before(*force_limited_until)))
		    continue;
		if (force_limited_since)
		  if (!(after = ymd.after(*force_limited_since)))
		    continue;
		  else {
		    ymd.front = *after;
		    year.front = ymd.front.year();
		  }
	      }
	      gregorian::date dt0(year.front, month, day), dt1 = dt0 + gregorian::date_duration(1);
	      ostringstream os;
	      os << format(vevent_fmt[0]) % to_iso_string(dt) %
		((summary != "") ? summary : default_summary[k]) %
		((location != "") ? location : default_location[k]) %
		((description != "") ? description : default_description[k]) %
		(format(vevent_fmt[2]) % to_iso_string(dt0)) % (format(vevent_fmt[2]) % to_iso_string(dt1)) %
		default_categories %
		default_transp;
	      if (year.is_repeat()) {
		if (force_limited_until)
		  if (before) {
		    ymd.back = *force_limited_until;
		    if (normalize_until)
		      ymd.back = *before;
		    year.back = ymd.back.year();
		  }
		os << format("RRULE:FREQ=YEARLY;INTERVAL=%1%;BYMONTHDAY=%2%") % year.diff % day;
		if (convert_until_count && !year.is_unlimited())
		  os << format(";COUNT=%1%") % year.count();
		else if (!force_unlimited_repeat && !year.is_unlimited())
		  os << format(";UNTIL=%1%T000000Z") %
		    (lexical_cast<string>(normalize_until ? year.last() : year.back) + "12" + "31");
		os << "\r\n";
	      }
	      string s = regex_replace(os.str(), regex_map["\\\\r\\\\n"], "\r\n");
	      cout << s << format(vevent_fmt[1]);
	    }
	  }
	}
	break;
      }
    }
  }
  cout << format(vcalendar_fmt[1]);
  return 0;
}
